import scilight as sl

# ------------------------------------------------------------------------
# Download a gzipped fasta file and save it as chrmt.fa.gz
# ------------------------------------------------------------------------
url = "ftp://ftp.ensembl.org/pub/release-100/fasta/homo_sapiens/dna/Homo_sapiens.GRCh38.dna.chromosome.MT.fa.gz"
download_task = sl.shell(f"wget -O [o:gz:chrmt.fa.gz] {url}")

# ------------------------------------------------------------------------
# Un-GZip the file, into a file named chrmt.fa
# ------------------------------------------------------------------------
ungzip_task = sl.shell(
    "zcat [i:gz] > [o:fa:[i:gz|%.gz]]", inputs={"gz": download_task.outputs["gz"]}
)

# ------------------------------------------------------------------------
# Count the fraction of G+C, vs G+C+A+T
# ------------------------------------------------------------------------
# A function for Count GC fraction in DNA
def count_gcfrac_func(task):
    gc_count = 0
    at_count = 0

    with open(task.inputs["fa"]) as infile:
        for line in infile:
            if line[0] == ">":
                continue
            for char in line:
                if char in ["A", "T"]:
                    at_count += 1
                elif char in ["G", "C"]:
                    gc_count += 1

    gc_fraction = gc_count / (gc_count + at_count)

    with open(task.outputs["gcfrac"], "w") as outfile:
        outfile.write(str(gc_fraction) + "\n")


# Execute the function
count_task = sl.func(
    count_gcfrac_func,
    inputs={"fa": ungzip_task.outputs["fa"]},
    outputs={"gcfrac": "gcfrac.txt"},
)



import subprocess as sp
from subprocess import CompletedProcess
import os
import os.path
import pathlib
import re
import shutil
import sys
from typing import Callable, Dict


class Task:
    """
    Super-class for tasks.
    """

    def __init__(
        self,
        inputs: Dict[str, str] = {},
        outputs: Dict[str, str] = {},
        params: Dict[str, str] = {},
        options: Dict[str, str] = {},
    ):
        self.inputs = inputs
        self.params = params

        self.outputs = {}
        for outname, outpath in outputs.items():
            outpath, _ = self._replace_placeholders(outpath)
            self.outputs[outname] = outpath

        self.tempfiles = True
        if "tempfiles" in options:
            self.tempfiles = options["tempfiles"]

    def execute(self) -> None:
        raise NotImplementedError("execute method not implemented")

    def _outputs_exist(self) -> bool:
        for name, path in self.outputs.items():
            if os.path.exists(path + ".tmp"):
                raise Exception("Existing temp files found: %s.tmp" % path)
            if os.path.exists(path):
                print(
                    "File or folder already exists, so skipping task: %s (%s)"
                    % (path, name)
                )
                return True
        return False

    def _ensure_output_folders_exist(self) -> None:
        for out_path in self.outputs.values():
            parent_dir = pathlib.Path(out_path).parent
            if not parent_dir.exists():
                os.makedirs(parent_dir)

    def _replace_placeholders(self, shell: str):
        # In-ports
        ms = re.findall(r"(\[i\:([^:\]\|]*)(\|([^\]]+))?\])", shell, flags=re.S)
        for m in ms:
            placeholder = m[0]
            portname = m[1]
            path = self.inputs[portname]

            if m[3] != "":
                modifiers = m[3]
                mods = modifiers.strip("|").split("|")
                for mod in mods:
                    # Replace extensions specified with |%.ext modifier
                    if mod[0] == "%":
                        extlen = len(mod[1:])
                        path = path[0:-extlen]
                    if mod[0] == "s":
                        mod_parts = re.match(r"s\/([^/]+)\/([^/]+)\/", mod, flags=re.S)
                        search = mod_parts[1]
                        replace = mod_parts[2]
                        path = path.replace(search, replace)
                    # Take basename of path, if |basename modifier is found
                    if mod == "basename":
                        path = os.path.basename(path)

            shell = shell.replace(placeholder, path)

        # Parameters
        ms = re.findall(r"(\[p\:([^:\]\|]*)(\|([^\]]+))?\])", shell, flags=re.S)
        for m in ms:
            placeholder = m[0]
            param_name = m[1]
            param_value = self.params[param_name]

            if m[3] != "":
                modifiers = m[3]
                mods = modifiers.strip("|").split("|")
                for mod in mods:
                    # Replace extensions specified with |%.ext modifier
                    if mod[0] == "s":
                        mod_parts = re.match(r"s\/([^/]+)\/([^/]+)\/", mod, flags=re.S)
                        search = mod_parts[1]
                        replace = mod_parts[2]
                        param_value = param_value.replace(search, replace)

            shell = shell.replace(placeholder, param_value)

        temp_shell = shell
        # Out-ports
        ms = re.findall(r"(\[o\:([^:\]]*)(:([^:\]]+))?\])", shell, flags=re.S)
        for m in ms:
            placeholder = m[0]
            portname = m[1]
            if m[3] != "":
                path = m[3]
                self.outputs[portname] = path
            else:
                path = self.outputs[portname]

            temppath = "%s.tmp" % path

            shell = shell.replace(placeholder, path)
            temp_shell = temp_shell.replace(placeholder, temppath)

        return shell, temp_shell

    def _move_tempfiles_to_final_path(self) -> None:
        for _, path in self.outputs.items():
            shutil.move("%s.tmp" % path, path)


class FuncTask(Task):
    def __init__(
        self,
        func,
        inputs: Dict[str, str] = {},
        outputs: Dict[str, str] = {},
        params: Dict[str, str] = {},
        options: Dict[str, str] = {},
    ):
        super(FuncTask, self).__init__(inputs, outputs, params, options)
        self.func = func

    def execute(self) -> None:
        if self._outputs_exist():
            return

        self._ensure_output_folders_exist()

        if self.tempfiles:
            # Make paths into temp paths
            for name, path in self.outputs.items():
                self.outputs[name] = path + ".tmp"

        print(
            "Executing python function, producing output(s): %s"
            % ", ".join(self.outputs.values())
        )
        self.func(self)

        if self.tempfiles:
            # Make paths into normal paths again
            for name, path in self.outputs.items():
                self.outputs[name] = path[:-4]

            self._move_tempfiles_to_final_path()


class ShellTask(Task):
    def __init__(
        self,
        command: str,
        inputs: Dict[str, str] = {},
        outputs: Dict[str, str] = {},
        params: Dict[str, str] = {},
        options: Dict[str, str] = {},
    ):
        super(ShellTask, self).__init__(inputs, outputs, params, options)
        self.command, self.temp_command = self._replace_placeholders(command)

    # ------------------------------------------------
    # Public methods
    # ------------------------------------------------
    def execute(self) -> None:
        if self._outputs_exist():
            return

        self._ensure_output_folders_exist()

        out = self._execute_shell_command_get_all_output(
            self.command, self.temp_command
        )
        self._add_cmd_results(out)
        if self.tempfiles:
            self._move_tempfiles_to_final_path()

    # ------------------------------------------------
    # Internal methods
    # ------------------------------------------------
    def _execute_shell_command_get_all_output(
        self, command: str, temp_command: str
    ) -> CompletedProcess:
        print("Executing command: %s" % command)
        cmd = command
        if self.tempfiles:
            cmd = temp_command
        out = sp.run(cmd, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
        if out.returncode != 0:
            print(f"ERROR: Command returned a non-zero return code: {cmd}")
            print("STDERR:")
            print(str(out.stderr).replace("\\n", "\n"))
            print("STDOUT:")
            print(str(out.stdout).replace("\\n", "\n"))
            sys.exit(1)
        return out

    def _add_cmd_results(self, cmdout: CompletedProcess):
        self.args = cmdout.args
        self.returncode = cmdout.returncode
        self.stdout = cmdout.stdout
        self.stderr = cmdout.stderr


def shell(
    command: str,
    inputs: Dict[str, str] = {},
    outputs: Dict[str, str] = {},
    params: Dict[str, str] = {},
    options: Dict[str, str] = {},
) -> ShellTask:
    """
    Return a new shell task configured by the shell command `command` and
    dictionaries for inputs and outputs (with names as keys and path formats as
    values), params and options.
    """
    task = ShellTask(command, inputs, outputs, params, options)
    task.execute()
    return task


def func(
    func: Callable[[FuncTask], None],
    inputs: Dict[str, str] = {},
    outputs: Dict[str, str] = {},
    params: Dict[str, str] = {},
    options: Dict[str, str] = {},
) -> FuncTask:
    task = FuncTask(func, inputs, outputs, params, options)
    task.execute()
    return task



import os
from pytest import fail
import sys
import shutil

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
import scilight as sl


def test_replace_placeholders():
    task = sl.ShellTask("", inputs={"gz": "data/chrmt.fa.gz"})
    for cmdpattern, expected_cmd, expected_temp_cmd in [
        ("wget -O [o:fasta:chrmt.fa]", "wget -O chrmt.fa", "wget -O chrmt.fa.tmp"),
        (
            "wget -O [o:fasta:[i:gz]]",
            "wget -O data/chrmt.fa.gz",
            "wget -O data/chrmt.fa.gz.tmp",
        ),
        (
            "zcat [i:gz] > [o:fasta:[i:gz|%.gz]]",
            "zcat data/chrmt.fa.gz > data/chrmt.fa",
            "zcat data/chrmt.fa.gz > data/chrmt.fa.tmp",
        ),
        (
            "zcat [i:gz] > [o:fasta:[i:gz|basename|%.gz]]",
            "zcat data/chrmt.fa.gz > chrmt.fa",
            "zcat data/chrmt.fa.gz > chrmt.fa.tmp",
        ),
        (
            "zcat [i:gz] > [o:fasta:[i:gz|%.gz|basename]]",
            "zcat data/chrmt.fa.gz > chrmt.fa",
            "zcat data/chrmt.fa.gz > chrmt.fa.tmp",
        ),
    ]:
        cmd, temp_cmd = task._replace_placeholders(cmdpattern)
        if cmd != expected_cmd:
            fail(
                'Expected command: "{expected}", but got "{actual}"'.format(
                    expected=expected_cmd, actual=cmd
                )
            )
        if temp_cmd != expected_temp_cmd:
            fail(
                'Expected temp command: "{expected}", but got "{actual}"'.format(
                    expected=expected_temp_cmd, actual=temp_cmd
                )
            )


def test_sl_shell():
    sl.shell(
        "wget -O [o:gz:/tmp/testdir/testsubdir/chrmt.fa.gz] ftp://ftp.ensembl.org/pub/release-100/fasta/homo_sapiens/dna/Homo_sapiens.GRCh38.dna.chromosome.MT.fa.gz"
    )
    if not os.path.isfile("/tmp/testdir/testsubdir/chrmt.fa.gz"):
        fail("download did not succeed")

    shutil.rmtree("/tmp/testdir")


def test_output_path_formatting():
    t1 = sl.shell(
        "echo hej > [o:hej:/tmp/hej.txt]"
    )
    t2 = sl.shell(
        "cat [i:hej] > [o:hejdaa]",
        inputs={"hej": t1.outputs["hej"]},
        outputs={"hejdaa": "[i:hej|%.txt|s/hej/hi/].daa.foo_[p:foo].txt"},
        params={"foo": "bar"}
    )
    expected_output_file = "/tmp/hi.daa.foo_bar.txt"
    if not os.path.isfile(expected_output_file):
        fail(f"Failed to create file {expected_output_file}!")

    os.remove(expected_output_file)


def test_sl_func():
    def write_file(task):
        with open(task.outputs["out"], "w") as outfile:
            outfile.write("test-output\n")

    sl.func(write_file, outputs={"out": "/tmp/output.txt"})

    if not os.path.isfile("/tmp/output.txt"):
        fail("writing of file in sl.func() did not succeed")

    os.remove("/tmp/output.txt")



import os
import sys

try:
    from setuptools import setup
except:
    from distutils.core import setup

with open("README.md") as fh:
    long_description = fh.read()

setup(
    name="scilight",
    version="0.6.0",
    description="Workflow library in pure python, for executing shell commands saving data to the file system without re-executing already executed tasks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Samuel Lampa",
    author_email="shl@rilspace.com",
    url="https://github.com/samuell/scilight",
    license="MIT",
    keywords="workflows workflow pipeline task",
    packages=[
        "scilight",
    ],
    install_requires=[],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
)
